import os
import lighter
from string import Template
from shutil import copyfile


def get_lighter_root():
    return '/'.join(os.path.abspath(lighter.__file__).split('/')[:-1])


def get_lighter_path(module_name, source_name):
    root = get_lighter_root()
    path = 'templates'
    if module_name is not None:
        path = os.path.join(path, module_name)
    path = os.path.join(path, source_name)
    template_file_name = os.path.join(root, path)
    return os.path.exists(template_file_name), template_file_name


def create_template_file(project_name, source_name, target_name, template, module_name=None, override=False):
    exists, template_file_name = get_lighter_path(module_name, source_name)
    if not exists:
        return
    with open(template_file_name, 'r') as file:
        src = Template(file.read())
        result = src.substitute(template)
    target_dir_name = project_name
    if module_name is not None:
        target_dir_name = os.path.join(project_name, module_name)
    if not os.path.exists(target_dir_name):
        os.makedirs(target_dir_name)
    target_file_name = os.path.join(target_dir_name, target_name)
    if override or not os.path.exists(target_file_name):
        with open(target_file_name, 'w') as file:
            file.write(result)


def copy_file(project_name, source_name, target_name, module_name=None, override=False):
    exists, template_file_name = get_lighter_path(module_name, source_name)
    if not exists:
        return
    target_dir_name = project_name
    if module_name is not None:
        target_dir_name = os.path.join(project_name, module_name)
    if not os.path.exists(target_dir_name):
        os.makedirs(target_dir_name)
    target_path = os.path.join(target_dir_name, target_name)
    if override or not os.path.exists(target_path):
        copyfile(template_file_name, target_path)


def create_init(project_name, module_name: str = None):
    if module_name is None:
        target_dir_name = project_name
    else:
        target_dir_name = os.path.join(project_name, module_name)
    if not os.path.exists(target_dir_name):
        os.makedirs(target_dir_name)
    target_file_name = os.path.join(target_dir_name, '__init__.py')
    with open(target_file_name, 'a'):
        pass



from threading import RLock
from box import Box


class Registry(object):
    """
    Instance and type registry for properties loaded from configs.
    """
    _instance = None
    # mutex for threadsafe usage
    _mutex = RLock()

    def __init__(self, **kwargs):
        """
        Creates a new Registry object storing and handling all types and instances imported
        from configs.
        :param kwargs:
        """
        super(Registry, self).__init__(**kwargs)
        self.instances = Box()
        self.types = Box()

    def register_type(self, name, type_):
        setattr(self.types, name, type_)

    def register_instance(self, name, instance):
        setattr(self.instances, name, instance)

    def unregister_type(self, name):
        del self.types[name]

    def unregister_instance(self, name):
        del self.instances[name]

    def contains_type(self, name):
        return name in self.types.keys()

    def contains_instance(self, name):
        return name in self.instances.keys()

    @staticmethod
    def get_instance() -> "Registry":
        Registry._mutex.acquire()
        try:
            return Registry._instance
        finally:
            Registry._mutex.release()

    @staticmethod
    def create_instance() -> "Registry":
        Registry._mutex.acquire()
        try:
            Registry._instance = Registry()
            return Registry._instance
        finally:
            Registry._mutex.release()



from lighter.decorator import context
from lighter.functional.metric import *


class BaseMetric(object):
    """
    Base metric class to register the required losses.
    """
    @context
    def __init__(self):
        pass

    def __call__(self, prediction, target):
        return {
            'acc': ACC(target, prediction).item(),
            'bacc': BACC(target, prediction).item(),
            'precision': precision(target, prediction).item(),
            'recall': recall(target, prediction).item(),
            'f1_score': f1_score(target, prediction).item(),
            'tpr': TPR(target, prediction).item(),
            'tnr': TNR(target, prediction).item(),
            'fpr': FPR(target, prediction).item(),
            'fnr': FNR(target, prediction).item()
        }



import numpy as np
from box import Box
from lighter.decorator import context


class BaseCollectible(object):
    """
    Base class for collectibles. They are used to store and update values over multiple steps / epochs.
    Collectibles are passed a dictionary containing key-value pairs whereas the key
    determines the collection for adding values.
    """
    @context
    def __init__(self):
        self.collection = Box()

    def reset(self):
        """
        Resets the collectible instance.
        :return:
        """
        self.collection = Box()

    def update(self, category: str = None, **kwargs):
        """
        Updates a collectible key-value pair.
        Dict: {
            'name': <number>,
            ...
        }
        :param category: defines a group category for the values
        :param kwargs: key-value data pair.
        :return:
        """
        for key, value in kwargs.items():
            if category is not None:
                key = '${}$_{}'.format(category, key)
            if key not in self.collection.keys():
                self.collection[key] = []
            self.collection[key].append(value)

    def redux(self, func=np.mean):
        """
        Reduces the collected values to a single number.
        :param func: Reduction function applied on the collected values.
        :return: Dictionary with reduced single numbers per key.
        """
        redux = {}
        for key, value in self.collection.items():
            redux[key] = func(np.array(value))
        return redux



def TP(target, prediction):
    """
    True positives.
    :param target: target value
    :param prediction: prediction value
    :return:
    """
    return (target.float() * prediction.float().round()).sum()


def TN(target, prediction):
    """
    True negatives.
    :param target: target value
    :param prediction: prediction value
    :return:
    """
    return ((target == 0).float() * (prediction.float().round() == 0).float()).sum()


def FP(target, prediction):
    """
    False positives.
    :param target: target value
    :param prediction: prediction value
    :return:
    """
    return ((target == 0).float() * prediction.float().round()).sum()


def FN(target, prediction):
    """
    False negatives.
    :param target: target value
    :param prediction: prediction value
    :return:
    """
    return (target.float() * (prediction.float().round() == 0).float()).sum()


def TPR(target, prediction, eps=1e-7):
    """
    True positive rate metric.
    :param target: target value
    :param prediction: prediction value
    :param eps: epsilon to avoid zero division
    :return:
    """
    tp = TP(target, prediction)
    fn = FN(target, prediction)
    s = tp + fn + eps
    return tp / s


def TNR(target, prediction, eps=1e-7):
    """
    True negative rate metric.
    :param target: target value
    :param prediction: prediction value
    :param eps: epsilon to avoid zero division
    :return:
    """
    tn = TN(target, prediction)
    fp = FP(target, prediction)
    s = (tn + fp + eps)
    assert (s > 0)
    return tn / s


def FPR(target, prediction, eps=1e-7):
    """
    False positive rate metric.
    :param target: target value
    :param prediction: prediction value
    :param eps: epsilon to avoid zero division
    :return:
    """
    fp = FP(target, prediction)
    tn = TN(target, prediction)
    s = fp + tn + eps
    assert (s > 0)
    return fp / s


def FNR(target, prediction, eps=1e-7):
    """
    False negative rate metric.
    :param target: target value
    :param prediction: prediction value
    :param eps: epsilon to avoid zero division
    :return:
    """
    fn = FN(target, prediction)
    tp = TP(target, prediction)
    s = fn + tp + eps
    assert (s > 0)
    return fn / s


def ACC(target, prediction, eps=1e-7):
    """
    Accuracy metric.
    :param target: target value
    :param prediction: prediction value
    :param eps: epsilon to avoid zero division
    :return:
    """
    tp = TP(target, prediction)
    tn = TN(target, prediction)
    p = target.sum().float()
    n = (target == 0).sum().float()
    s = p + n + eps
    assert (s > 0)
    return (tp + tn) / s


def BACC(target, prediction, eps=1e-7):
    """
    Balanced accuracy metric.
    :param target: target value
    :param prediction: prediction value
    :param eps: epsilon to avoid zero division
    :return:
    """
    tp = TP(target, prediction)
    tn = TN(target, prediction)
    p = target.sum().float()
    n = (target == 0).sum().float()
    s = (p + eps) + tn / (n + eps)
    assert (s > 0)
    return 0.5 * (tp / s)


def precision(target, prediction, eps=1e-7):
    """
    Precision metric.
    :param target: target value
    :param prediction: prediction value
    :param eps: epsilon to avoid zero division
    :return:
    """
    tp = TP(target, prediction)
    fp = FP(target, prediction)
    s = (tp + fp + eps)
    assert (s > 0)
    return tp / s


def recall(target, prediction, eps=1e-7):
    """
    Recall metric.
    :param target: target value
    :param prediction: prediction value
    :param eps: epsilon to avoid zero division
    :return:
    """
    tp = TP(target, prediction)
    fn = FN(target, prediction)
    s = (tp + fn + eps)
    assert (s > 0)
    return tp / s


def f1_score(target, prediction, eps=1e-7):
    """
    F1-score metric.
    :param target: target value
    :param prediction: prediction value
    :param eps: epsilon to avoid zero division
    :return:
    """
    precision_ = precision(target, prediction)
    recall_ = recall(target, prediction)
    n = (precision_ * recall_)
    s = (precision_ + recall_ + eps)
    assert (s > 0)
    return 2 * (n / s)



import json
import os
import argparse
from threading import RLock
import logging
import torch
from box import Box
from lighter.exceptions import InvalidTypeReferenceError
from lighter.loader import Loader
from lighter.misc import extract_named_args, try_to_number_or_bool, DotDict
from lighter.registry import Registry


def import_value_rec(name, value):
    """
    Imports configs or classes to the current config instance.
    :param name: property name
    :param value: property value
    :return: value instance of the replaced property
    """
    # import classes if type:: reference was found in json
    if isinstance(value, str) and "type::" in value:
        try:
            type_ = Loader.import_path(value[len("type::"):])
            if type_ is None:
                raise InvalidTypeReferenceError('Config: Could not find specified reference: {}'.format(value))

            # register type to registry
            Registry.get_instance().register_type(name, type_)

        except ModuleNotFoundError as e:
            logging.warning("Error while importing '{}' - {}".format(value, e))
    # import config if config:: reference was found in json
    elif isinstance(value, str) and "config::" in value:
        try:
            conf = Config(path=value[len("config::"):])
            conf = Box(conf)
            conf['config_path'] = value[len("config::"):]
            value = conf
        except ModuleNotFoundError as e:
            logging.warning("Error while importing '{}' - {}".format(value, e))
    # else regular set
    elif isinstance(value, dict):
        for k, v in value.items():
            value[k] = import_value_rec(k, v)
        value = Box(value)

    return value


def override(parent, name, value):
    """
    Overrides a config key.
    :param parent: Parent dictionary instance
    :param name: name of the property
    :param value: value for the property
    :return:
    """
    if value is not None:
        # import key-values into the current property instance
        if isinstance(value, str) and "import::" in value:
            conf = Config(path=value[len("import::"):])
            for k, v in Box(conf).items():
                setattr(parent, k, v)
                logging.info("Config: {}={}".format(k, getattr(parent, k)))
        # decent down the dictionary instance
        elif isinstance(value, dict):
            dict_ = Box(value)
            for k, v in dict_.copy().items():
                override(dict_, k, v)
                # remove old import:: reference after import
                if isinstance(v, str) and "import::" in v:
                    del dict_[k]
            setattr(parent, name, dict_)
        else:
            value = import_value_rec(name, value)
            setattr(parent, name, value)
            logging.info("Config: {}={}".format(name, getattr(parent, name)))


class Config(Box):
    _instance = None
    # thread save lock to ensure single instance creation for the default config
    _mutex = RLock()

    def __init__(self, path: str = None, override_args=None, **kwargs):
        """
        Create config object from json file.

        :param path: optional;
            If passed read config from specified file, otherwise parse command line for config parameter and optionally
            override arguments.
        :param override_args: args to override the current properties
        """
        super(Config, self).__init__()
        if kwargs:
            self.initialize_from_json(kwargs.items())
        # Read config and override with args if passed
        if path is not None:
            if not os.path.exists(path):
                raise FileNotFoundError('Invalid config path! Could not resolve: {}'.format(path))
            with open(path) as f:
                self.initialize_from_json(json.loads(f.read()).items())
            # override if necessary
            if override_args is not None:
                self.override_from_commandline(override_args)

    def copy(self) -> "Config":
        config = Config()
        dict_ = super().copy()
        config.initialize_from_dict(dict_)
        return config

    def save(self, file_name):
        dict_str = json.dumps(self, indent=2)
        with open(file_name, 'w') as file:
            file.write(dict_str)

    def set_value(self, name, value):
        """Sets the properties recursively according to a dot-separated reference.
        """
        parent, name = DotDict.resolve(self, name)
        override(parent, name, value)

    def has_value(self, name):
        """Checks properties recursively according to a dot-separated reference.
        """
        parent, name = DotDict.resolve(self, name)
        return hasattr(parent, name)

    def get_value(self, name, default=None):
        """Returns the properties recursively according to a dot-separated reference.
        """
        parent, name = DotDict.resolve(self, name)
        return getattr(parent, name, default)

    def initialize_from_dict(self, kv_pairs):
        if kv_pairs:
            for i, (name, value) in enumerate(kv_pairs.items()):
                override(self, name, value)

    def initialize_from_json(self, nv_pairs=None):
        if nv_pairs:
            for i, (name, value) in enumerate(nv_pairs):
                override(self, name, value)

    def override_from_commandline(self, override_args=None):
        if override_args is not None:
            override_dict = extract_named_args(override_args)
            for k, v in override_dict.items():
                name = k[2:] if "--" in k else k  # remove leading --
                if v is None:
                    value = True  # assume cmd switch
                else:
                    value = v if v.startswith('"') or v.startswith("'") else try_to_number_or_bool(v)
                if "." in name:
                    names = name.split(".")
                    name = names[0]
                    if len(names) == 2:
                        if hasattr(self, names[0]):
                            curdict = getattr(self, names[0])
                        else:
                            curdict = dict()
                        curdict[names[1]] = value
                        value = curdict
                    elif len(names) == 3:
                        if hasattr(self, names[0]):
                            curdict = getattr(self, names[0])
                        else:
                            curdict = dict()

                        if names[1] in curdict:
                            subdict = curdict[names[1]]
                        else:
                            curdict[names[1]] = dict()
                            subdict = curdict[names[1]]

                        subdict[names[2]] = value
                        value = curdict
                    else:
                        raise Exception("Unsupported command line option (can only override dicts with 1 or 2 levels)")
                override(self, name, value)

    @staticmethod
    def create_instance(config_file: str = None,
                        config_dict: dict = None,
                        parse_args_override: bool = True,
                        device: str = None) -> "Config":
        """
        Create default config instance. Loads the command line overridable config settings and
        sets the default device instance.
        :param config_file: path to config file
        :param config_dict: pre-configs
        :param parse_args_override: override
        :param device: training device
        :return:
        """
        if config_dict is None:
            config_dict = {}
        Config._mutex.acquire()
        try:
            Config._instance, args = Config.load(path=config_file,
                                                 parse_args_override=parse_args_override,
                                                 **config_dict)
            # load default device
            if args is not None and args.device is not None:
                Config._instance.set_value('device.default', args.device)
            elif device is not None:
                Config._instance.set_value('device.default', device)
            elif torch.cuda.is_available():
                Config._instance.set_value('device.default', 'cuda')
            else:
                Config._instance.set_value('device.default', 'cpu')
            return Config._instance
        finally:
            Config._mutex.release()

    @staticmethod
    def get_instance() -> "Config":
        """
        Return default config instance.
        :return:
        """
        Config._mutex.acquire()
        try:
            return Config._instance
        finally:
            Config._mutex.release()

    @staticmethod
    def load(path: str = None, parse_args_override: bool = False, **kwargs):
        """
        Loads a property file and allows to override arguments based on command line overrides.
        :param path: path to property file
        :param parse_args_override: fallback to --config if no path was specified.
        :return: Config instance or Config with command line args tuple if parse_args_fallback is enabled
        """
        if parse_args_override:
            parser = argparse.ArgumentParser()
            parser.add_argument('--config', type=str, help='path to config file')
            parser.add_argument('--device', type=str, help='set the used device')
            args = parser.parse_args()
            if args.config is not None:
                path = args.config
            return Config(path=path, **kwargs), args
        return Config(path=path, **kwargs), None



import traceback
from threading import RLock

import logging

from lighter.config import Config
from lighter.exceptions import TypeInstantiationError
from lighter.misc import generate_short_id
from lighter.registry import Registry
from lighter.search import ParameterSearch

import numpy as np

# TODO: remove in future terms
import warnings
warnings.simplefilter('ignore', FutureWarning)


class Context(object):
    """
    Application context handling all the config, registry and search parameter instances.
    """
    # mutex to protect context from multiprocess concurrent access
    _mutex = RLock()
    _instance = None

    def __init__(self,
                 config_file: str,
                 config_dict: dict,
                 parse_args_override,
                 device: str = None,
                 auto_instantiate_types: bool = True,
                 allow_context_changes: bool = True):
        """
        Create a context object and registers configs, types and instances of the defined modules.
        :param config_file: default config file
        """
        Context._mutex.acquire()
        try:
            self.registry = Registry.create_instance()
            self.config = Config.create_instance(config_file,
                                                 config_dict=config_dict,
                                                 parse_args_override=parse_args_override,
                                                 device=device)
            self.config.set_value('context_id', generate_short_id())
            self.search = ParameterSearch.create_instance()
            self.allow_context_changes = allow_context_changes
            self.auto_instantiate_types = auto_instantiate_types
            if auto_instantiate_types:
                # if auto instantiation of types is enables, and no context is available this requires to assign
                # the current instance to the global instance
                if Context._instance is None:
                    Context._instance = self
                self.instantiate_types()
        finally:
            Context._mutex.release()

    def _resolve_dependency_graph(self, types: dict, num_retries: int = 0, max_retries: int = np.inf, error=[]):
        """
        Registers instances with retries if graph resolving fails due to transitive dependencies between
        modules.
        TODO: improve by different prioritization technique
        :param types: types to instantiate
        :param num_retries: current retry count
        :param max_retries: maximum recursion depth
        :return:
        """
        retry = {}
        for name, class_ in types.items():
            try:
                self.registry.register_instance(name, class_())
            except Exception:
                err = traceback.format_exc()
                error.append(err)
                retry[name] = class_
        if len(retry) > 0 and num_retries < max_retries:
            num_retries += 1
            self._resolve_dependency_graph(retry, num_retries, max_retries)
        if num_retries >= max_retries:
            for e in error:
                logging.exception(e, exc_info=True)
            raise TypeInstantiationError('Could not resolve the dependency graph due to type creation errors.')

    def instantiate_types(self, types: dict = None):
        """
        Instantiates and registers types to the registry instances.
        :param types: types to instantiate and register
        :return:
        """
        Context._mutex.acquire()
        if types is None:
            types = self.registry.types
        try:
            len_types = len(list(types.keys()))
            if len_types > 0:
                self._resolve_dependency_graph(types, 0, max_retries=len_types)
        finally:
            Context._mutex.release()

    @staticmethod
    def get_instance() -> "Context":
        Context._mutex.acquire()
        try:
            return Context._instance
        finally:
            Context._mutex.release()

    @staticmethod
    def create(config_file: str = None,
               config_dict: dict = None,
               parse_args_override: bool = True,
               device: str = None,
               auto_instantiate_types: bool = True,
               allow_context_changes: bool = True) -> "Context":
        """
        Create application context threadsafe.
        :param config_file: Initial config file for context to load.
        :param config_dict: Initial config json for context to load.
        :param parse_args_override: Allows to override configs according to the command line arguments.
        :param auto_instantiate_types: Instantiates types after been registered
        :param device: specifies the running device
        :param allow_context_changes: If this is set to False, it prevents all decorators from modifying updates
               on the configs, context and registry
        :return:
        """
        Context._mutex.acquire()
        try:
            Context._instance = Context(config_file,
                                        config_dict=config_dict,
                                        parse_args_override=parse_args_override,
                                        auto_instantiate_types=auto_instantiate_types,
                                        device=device,
                                        allow_context_changes=allow_context_changes)
            return Context._instance
        finally:
            Context._mutex.release()



from torch.utils.data import Dataset
from lighter.decorator import transform


class BaseDataset(Dataset):
    """
    Dataset base class preparing the data.
    """
    @transform
    def __init__(self):
        pass



import json
import petname
import sys
from box import Box


def extract_named_args(arg_list):
    result = {}
    for i in range(0, len(arg_list)):
        a = arg_list[i]
        if a.startswith("--"):
            if i + 1 < len(arg_list) and not arg_list[i + 1].startswith("--"):
                result[a] = arg_list[i + 1]
            else:
                result[a] = None
    return result


def try_to_number_or_bool(value):
    try:
        value = int(value)
    except ValueError:
        try:
            value = float(value)
        except ValueError:
            if isinstance(value, str) and (value.lower() == "false" or value.lower() == "true"):
                value = (value.lower() == "true")
            pass
    return value


def generate_short_id():
    return petname.Generate(2, '-', 6)


def generate_long_id():
    return petname.Generate(3, '-', 6)


def _to_json(obj, filename=None,
             encoding="utf-8", errors="strict", **json_kwargs):
    json_dump = json.dumps(obj,
                           ensure_ascii=False, **json_kwargs)
    if filename:
        with open(filename, 'w', encoding=encoding, errors=errors) as f:
            f.write(json_dump if sys.version_info >= (3, 0) else
                    json_dump.decode("utf-8"))
    else:
        return json_dump


class DotDict:
    @staticmethod
    def resolve(parent_ori, name_ori):
        parent = parent_ori
        prev_parent = parent_ori
        groups = name_ori.split('.')
        for group in groups[:-1]:
            parent = parent.get(group)
            if parent is None:
                parent = Box()
                setattr(prev_parent, group, parent)
            else:
                prev_parent = parent
        return parent, groups[-1]



import os
import torch
import numpy as np
from tqdm import tqdm
from datetime import datetime
from lighter.decorator import device, context
from lighter.misc import generate_long_id


class BaseExperiment(object):
    """
    Experiment base class to create new algorithm runs.
    An experiment is also iterable and can be used as an iterator.
    It allows different hooks to interact in the main loop.
    """
    @device
    @context
    def __init__(self,
                 experiment_id: str = None,
                 epochs: int = 100,
                 enable_checkpoints: bool = True,
                 checkpoints_dir: str = 'runs/',
                 checkpoints_interval: int = 1):
        if experiment_id is None:
            experiment_id = generate_long_id()
        self.config['experiment_id'] = experiment_id
        self.epoch = 0
        self.epochs = epochs
        self.enable_checkpoints = enable_checkpoints
        self.checkpoints_dir = checkpoints_dir
        self.checkpoints_interval = checkpoints_interval

    def __call__(self, *args, **kwargs):
        self.run()

    def __iter__(self):
        self.initialize()
        self.eval()
        return self

    def __next__(self):
        if self.epoch < self.epochs:
            self._epoch(self.epoch)
            self.epoch += 1
            return self.epoch
        else:
            self.finalize()
            raise StopIteration

    def _epoch(self, epoch):
        self.pre_epoch()
        self.train()
        self.eval()
        self.checkpoint(epoch)
        self.post_epoch()

    def run(self):
        """
        Main entrance point for an experiment.
        :return:
        """
        self.initialize()
        self.eval()
        for epoch in tqdm(range(self.epochs)):
            self._epoch(epoch)
        self.finalize()

    def initialize(self):
        """
        Initialize the experiment phase.
        :return:
        """
        self.epoch = 0
        # save the new experiment config
        path = os.path.join(self.checkpoints_dir, self.config.context_id, self.config.experiment_id)
        if not os.path.exists(path):
            os.makedirs(path)
        config_file = os.path.join(path, 'experiment.config.json')
        self.config.save(config_file)

    def pre_epoch(self):
        """
        Hook that can be overridden before a training epoch starts.
        :return:
        """
        raise NotImplementedError('BaseExperiment: No implementation found!')

    def post_epoch(self):
        """
        Hock that can be overridden after training epoch ended.
        The current implementation resets the collectible and steps the writer.
        :return:
        """
        raise NotImplementedError('BaseExperiment: No implementation found!')

    def train(self):
        """
        Training instance for the experiment run.
        :return:
        """
        raise NotImplementedError('BaseExperiment: No implementation found!')

    def eval(self):
        """
        Evaluates an experiment.
        :return:
        """
        raise NotImplementedError('BaseExperiment: No implementation found!')

    def checkpoint(self, epoch: int):
        """
        Code for model state save.
        :param epoch: Current epoch executed.
        :return:
        """
        raise NotImplementedError('BaseExperiment: No implementation found!')

    def finalize(self):
        """
        Post experiment cleanup code.
        :return:
        """
        pass


class DefaultExperiment(BaseExperiment):
    """
    Simple implementation of the experiment base class to execute train / eval runs.
    """
    @device
    @context
    def __init__(self,
                 experiment_id: str = None,
                 epochs: int = 100,
                 enable_checkpoints: bool = True,
                 checkpoints_dir: str = 'runs/',
                 checkpoints_interval: int = 1,
                 redux_function=np.mean):
        super(DefaultExperiment, self).__init__(experiment_id=experiment_id,
                                                epochs=epochs,
                                                enable_checkpoints=enable_checkpoints,
                                                checkpoints_dir=checkpoints_dir,
                                                checkpoints_interval=checkpoints_interval)
        self.train_loader, self.val_loader = None, None
        self.redux_function = redux_function

    def initialize(self):
        # get data loaders
        self.train_loader, self.val_loader = self.data_builder.loader()
        super().initialize()

    def pre_epoch(self):
        pass

    def post_epoch(self):
        self.writer.step()
        self.collectible.reset()

    def train_batch(self, x, y):
        x, y = x.to(self.device), y.to(self.device)
        self.optimizer.zero_grad()
        pred = self.model(x)
        loss = self.criterion(pred, y)
        loss.backward()
        self.optimizer.step()
        self.collectible.update(category='train', **{'loss': loss.detach().cpu().item()})
        self.collectible.update(category='train', **self.metric(pred.detach().cpu(),
                                                                y.detach().cpu()))

    def train(self):
        if self.train_loader is not None:
            self.model.train()
            for i, (x, y) in enumerate(self.train_loader):
                self.train_batch(x, y)
            collection = self.collectible.redux(func=self.redux_function)
            self.writer.write(category='train', **collection)

    def eval_batch(self, x, y):
        x, y = x.to(self.device), y.to(self.device)
        pred = self.model(x)
        loss = self.criterion(pred, y)
        self.collectible.update(category='eval', **{'loss': loss.detach().cpu().item()})
        self.collectible.update(category='eval', **self.metric(pred.detach().cpu(),
                                                               y.detach().cpu()))

    def eval(self):
        if self.val_loader is not None:
            self.model.eval()
            with torch.no_grad():
                for i, (x, y) in enumerate(self.val_loader):
                    self.eval_batch(x, y)
                collection = self.collectible.redux(func=self.redux_function)
                self.writer.write(category='eval', **collection)

    def checkpoint(self, epoch: int):
        if self.enable_checkpoints and epoch % self.checkpoints_interval == 0:
            collection = self.collectible.redux(func=self.redux_function)
            timestamp = datetime.timestamp(datetime.now())
            file_name = 'e-{}_time-{}'.format(epoch, timestamp)
            path = os.path.join(self.checkpoints_dir, self.config.context_id, self.config.experiment_id)
            ckpt_file = os.path.join(path, '{}.ckpt'.format(file_name))
            torch.save({
                'epoch': epoch,
                'model_state_dict': self.model.state_dict(),
                'optimizer_state_dict': self.optimizer.state_dict(),
                'metrics': collection
            }, ckpt_file)



import zipfile, os
import urllib.request as request
import shutil


def download_and_extract_zip(url, dest_path):
    """
    Downloads and extracts a zip-file to the destination path.
    :param url: url to download zip file.
    :param dest_path: destination path to extract the files
    :return:
    """
    tmp_path = 'tmp'
    if not os.path.exists(tmp_path):
        os.makedirs(tmp_path)
    data, _ = request.urlretrieve(url, filename=os.path.join(tmp_path, 'file.zip'))
    with open(data, 'rb') as file:
        zf = zipfile.ZipFile(file)
        zf.extractall(path=dest_path)
    shutil.rmtree(tmp_path)



import torch

from lighter.decorator import context


class BaseCriterion(torch.nn.Module):
    """
    Criterion base class with injected application 'context'.
    """
    @context
    def __init__(self):
        super(BaseCriterion, self).__init__()



from lighter.decorator import strategy, references
from lighter.experiment import DefaultExperiment


class Experiment(DefaultExperiment):
    @strategy(config='configs/mnist.modules.config.json')
    @references
    def __init__(self):
        super(Experiment, self).__init__(epochs=30)
